package ch.hepia.tpscala.result


sealed trait Result[+E,+A] {

  def map[B]( f: A=>B ): Result[E,B] = {
    this match {
      case OK(a) => OK(f(a))
      case Err(e) => Err(e)
    }
  }

  //F >: E, signifie "F est un supertype de E"
  def flatMap[B, F >: E]( f: A=>Result[F,B] ): Result[F,B] = {
    this match {
      case OK(a) => f(a)
      case Err(e) => Err(e)
    }
  }

  //Retourne le résultat si OK
  //transforme l'erreur en une valeur valide sinon
  //B >: A signifie "B est un supertype de A
  //Une seul implémentation possible !
  def get[B >: A]( f: E => B ): B = {
    this match {
      case OK(a) => a
      case Err(e) => f(e)
    }
  }

  //Une seul implémentation possible !
  def fold[B]( f: E => B, g: A => B ): B = {
    this match {
      case OK(a) => g(a)
      case Err(e) => f(e)
    }
  }

  def toOption: Option[A] = {
    this match {
      case OK(a) => Some(a)
      case Err(_) => None
    }
  }

}
case class OK[A]( value: A ) extends Result[Nothing,A]
case class Err[E]( value: E ) extends Result[E,Nothing]
object Result {
  //Si il y a au moins une erreur dans la liste, le résultat est une erreur
  //Si il n'y a que des succès, liste les succès
  def sequence[E,A]( res: List[Result[E,A]] ): Result[E,List[A]] = {
    def seqRec(res: List[Result[E,A]], acc: List[A]): Result[E,List[A]] = {
      res match {
        case Nil => OK(acc.reverse)
        case OK(a) :: tl => seqRec(tl, a :: acc)
        case Err(e) :: _ => Err(e)
      }
    }
    seqRec(res, Nil)
  }
}


case class Config( hostname: String, port: Int )
object Config {
  private def line2keyValue( line: String ): Result[String,(String,String)] = {
    val elements = line.split("=")
    if( elements.size != 2 ) Err("Syntax error: " + line )
    else OK( (elements(0), elements(1) ) )
  }

  private def lines2map( lines: List[String] ): Result[String,Map[String,String]] = 
    Result.sequence( lines.map(line2keyValue) ).map( _.toMap )

  private def safeInt(str: String): Result[String, Int] = {
    try {
      OK(str.toInt)
    } catch {
      case _:Throwable => Err("Invalid port")
    }
  }

  private def mGet[K,V](mkv: Map[K,V], key: K): Result[String, V] = {
    mkv.get(key) match {
      case Some(value) => OK(value)
      case None => Err(key.toString + " not found")
    }
  }

  //Utilisez lines2map pour parvenir au résultat
  //Attention au clés manquantes
  def parse( lines: List[String] ): Result[String,Config] = {
    for {
      keyVal <- lines2map(lines)
      h <- mGet(keyVal, "hostname")
      pStr <- mGet(keyVal, "port")
      p <- safeInt(pStr)
    } yield Config(h,p)
  }
}
